/*
	MID/BID Database Utility for XFBB 7.01 or newer
	Copyright (c) 2005 by Stewart Wilkinson (G0LGS) 
*/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <fcntl.h>
#include <stdarg.h>
#include <string.h> 
#include <time.h>
#include <ctype.h>
#include <getopt.h>
#include <errno.h>
#include <unistd.h>
#include <sys/signal.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "utils.h"

extern char *optarg;
extern int optind;
extern int opterr;
extern int getopt(int argc, char * const *argv, const char *optstring);

char data_path[128];
char conf_path[128];
char dirmes_sys[128];
char dirmes_new[128];
char dirmes_old[128];
char wf_sys[128];
char wf_new[128];
char wf_old[128];
char s_string[13];

int  verbose = 0;
int  list = 0;
int  rebuild = 0;
int  search = 0;
long match = 0L;
long count = 0L;
long good = 0L;

#define _read read
#define _write write
#define O_BINARY 0
#define __a2__ __attribute__ ((packed, aligned(2)))

#define NBBBS 80
#define NBMASK NBBBS / 8

#define F_EXC	0x0001
#define F_LOC	0x0002
#define F_EXP	0x0004
#define F_SYS	0x0008
#define F_BBS	0x0010
#define F_PAG	0x0020
#define F_GST	0x0040
#define F_MOD	0x0080
#define F_PRV	0x0100
#define F_UNP	0x0200
#define F_NEW	0x0400
#define F_PMS	0x0800

#define EXC(flag)	(flag & F_EXC)
#define LOC(flag)	(flag & F_LOC)
#define EXP(flag)	(flag & F_EXP)
#define SYS(flag)	(flag & F_SYS)
#define BBS(flag)	(flag & F_BBS)
#define PAG(flag)	(flag & F_PAG)
#define GST(flag)	(flag & F_GST)
#define MOD(flag)	(flag & F_MOD)
#define PRV(flag)	(flag & F_PRV)
#define UNP(flag)	(flag & F_UNP)
#define NEW(flag)	(flag & F_NEW)
#define PMS(flag)	(flag & F_PMS)

typedef struct
{					/* = 194 bytes */
	char type;
	char status;
	long numero __a2__;
	long taille __a2__;
	long date __a2__;
	char bbsf[7];
	char bbsv[41];
	char exped[7];
	char desti[7];
	char bid[13];
	char titre[61];
	char bin;
	char free[9];
	unsigned short nblu;
	long theme __a2__;
	long datesd __a2__;
	long datech __a2__;
	char fbbs[NBMASK];
	char forw[NBMASK];
}
bullist;

typedef struct
{
	char	mode; 		/* Type of Message (A,B,P,T) */
	char	fbid[13];	/* BID of Message */
	long	msg_number __a2__;	/* Number of the Message */
}
bidfwd;

#define FBBCONF "/etc/ax25/fbb.conf"

int ReadInit()
{
	int len, i;
	int end = 0;
	char buffer[256];
	char *scan;
	char *ptr;
	char key[20];
	char pathname[256];
	FILE *init_ptr;

	if( (ptr = getenv("FBBCONF")) == NULL ){
		strcpy( pathname, FBBCONF);
	}else{
		strcpy( pathname, ptr);
	}

	if( access( pathname, R_OK ) != 0 ){
		fprintf( stderr, "Cannot find %s\n", pathname );
		return 1;
	}

	if ((init_ptr = fopen (pathname, "r")) != NULL) {

		if( verbose ){
			fprintf( stdout, "Reading %s.\n", pathname );
		}

		while( !end && fgets(buffer, sizeof(buffer), init_ptr) ){

			scan = buffer;

			/* Eat Spaces */
			while (isspace(*scan))
				++scan;

			/* Comments & Blanks ignored */
			if (*scan == '#' || *scan == '\0' )
				continue;

			/* remove leading \n */
			len = strlen(scan);
			if( len && (scan[len-1] == '\n'))
				scan[len-1] = '\0';

			/* keyword */
			i=0;
			ptr = key;
			while( isgraph(*scan)) {
				if( i++ < 4 ) {
					*ptr++ = *scan;
				}
				++scan;
				if(*scan == '=')
					break;         
			}
			*ptr = '\0';

			/* find the '=' */
			scan = strchr(scan, '=');

			/* copy value */
			if(scan) {
				++scan;
				while( isspace(*scan))
					++scan;

				/* if 1st char is '"' then string */
				if( *scan == '"' ){
					++scan;
					ptr = strrchr(scan,'"');
					if(ptr)
						*ptr = '\0';
				}

				if( strcmp( key, "data" ) == 0 ) {
					strcpy( data_path, scan );

				}

				if( strcmp( key, "conf" ) == 0 ) {
					strcpy( conf_path, scan );

				}

			}
		}

		fclose(init_ptr);
  
	}else{
		fprintf( stderr, "Unable to open '%s'\n", pathname );
		return 1;
	}

	return 0;
}

void do_head(char *cmd)
{
	fprintf( stdout, "\nXFBB %s V" Version " " Vdate " (c) by Stewart Wilkinson (G0LGS)\n\n", strupr( cmd ) );
}

void do_usage(char *cmd)
{
	do_head( cmd );

	if( strcasecmp( cmd, "bid-tool" ) == 0 ) {

		fprintf( stderr, "\tUsage: %s [options] | <MID/BID>\n\n", strupr( cmd) );
		fprintf( stderr, "\t--list\t\t\t\tList MID/BID and message numbers\n" );
		fprintf( stderr, "\t--search <MID/BID>\tSearch for MID/BID\n" );
		fprintf( stderr, "\t--rebuild\t\tRe-build the MID/BID Database from 'DIRMES.SYS'\n" );
		fprintf( stderr, "\t--verbose\t\tShow more information.\n" );
		fprintf( stderr, "\t-v | --version\n" );

	}else{

		fprintf( stderr, "\n\tUsage: %s\n\n", strupr(cmd) );
	}		

	fprintf( stderr, "\n" );
	
}


int Is_fbb(void) {

	int pid;

	pid = WEXITSTATUS(system("pidof xfbb >/dev/null 2>&1") );

	if( (pid != -1) && (pid != 0) ){
		pid = WEXITSTATUS( system("pidof xfbbd >/dev/null 2>&1") );

	}

	return(pid);
}


int do_bidlist(char *wf, char *mid)
{
	bidfwd bidrec;
	long last;
	FILE *wfptr;

	count = 0L;
	good = 0L;

	if( access( wf, R_OK ) != 0 ){
		fprintf( stderr, "unable to find %s\n", wf );
		return 1;
	}

	/* Open */
 	wfptr = fopen( wf, "rb");
	if ( wfptr == NULL ) {
		fprintf( stderr, "Unable to open '%s'\n", wf );
		return 1; 
	}

	/* Ignore 1st Record */
	fread(&bidrec, sizeof(bidfwd), 1, wfptr);

	last = bidrec.msg_number ;

	while( fread(&bidrec, sizeof(bidfwd), 1, wfptr) && ! feof( wfptr ) ){

		count++;
		if( (bidrec.mode != '\0') && (*bidrec.fbid != '\0') ){

			good++;

			if( mid == NULL ){
				fprintf( stdout, "$:%-13s ", bidrec.fbid );
				fprintf( stdout, "#:%ld ", bidrec.msg_number );
				fprintf( stdout, "T:%-c\n", bidrec.mode );

			}else if( search & (strstr( bidrec.fbid, mid ) != NULL) ){
				match ++;
				fprintf( stdout, "$:%-13s ", bidrec.fbid );
				fprintf( stdout, "#:%ld ", bidrec.msg_number );
				fprintf( stdout, "T:%-c\n", bidrec.mode );

			}else if( ! search & (strcmp( mid, bidrec.fbid ) == 0) ){
				match ++;
				fprintf( stdout, "$:%-13s ", bidrec.fbid );
				fprintf( stdout, "#:%ld ", bidrec.msg_number );
				fprintf( stdout, "T:%-c\n", bidrec.mode );
			}
		}
	}

	fclose( wfptr );

	return 0;

}

int rebuild_wfbids(void)
{
	bidfwd bidrec;
	bullist bullrec;
	FILE *dirsys;
	FILE *wfptr;

	if( ! verbose ){
		fprintf( stdout, "Processing " );
	}

	if( access( dirmes_sys, R_OK ) != 0 ){
		fprintf( stderr, "unable to find %s\n", dirmes_sys );
		return 1;
	}

	if( ! copy_ ( wf_sys, wf_old ) ){
		fprintf( stderr, "\nERROR: Unable to copy '%s' to '%s'\n", wf_sys, wf_old );
		return 1;
	}

	/* Open */
 	wfptr = fopen( wf_new, "wb");
	if ( wfptr == NULL ) {
		fprintf( stderr, "Unable to open '%s'\n", wf_new );
		return 1; 
	}

	/* Blank 1st Record for now */
	fseek(wfptr, sizeof(bidfwd), SEEK_SET);

	/* Open dirmes.sys */
	dirsys = fopen( dirmes_sys, "rb");
	if (dirsys == NULL) {
		fprintf( stderr, "Unable to open '%s'\n", dirmes_sys );
		return 1;
	}

	/* Skip the First Record - as it seems to be a duplicate of last message number */
	fseek( dirsys, sizeof(bullist), SEEK_SET );


	/* Read each record dirmes.sys and create wfbid.new */	
        count = 0;
        while( fread(&bullrec, sizeof(bullist), 1, dirsys)  && ! feof( dirsys ) ){

                count++;
                if( (! (count % 50)) && ( ! verbose) ){
                        fprintf( stdout, ".");
                        fflush( stdout );
                }

		if( verbose ){
			fprintf( stdout, "\rProcessing: %ld", count );
                        fflush( stdout );
		}

		/* Copy the Record to WFBID */
		strcpy( &bidrec.mode, &bullrec.type);
		strcpy( bidrec.fbid, bullrec.bid);
		bidrec.msg_number = bullrec.numero;

		/* Write the record */
		if( ( fwrite( &bidrec, sizeof(bidfwd), 1, wfptr ) == 0 ) ) {
			fprintf( stderr, "\nError: writing to %s\n", wf_new );
			fclose( wfptr );
			fclose( dirsys );
			return 1;
		}
        }

	fprintf( stdout, "\n");
        /* Finished reading dirmes.sys so close it */
        fclose( dirsys );

	/* Move to 1st record - to write Last Number */
	fseek(wfptr, 0, SEEK_SET);

	bidrec.msg_number = count;

	/* Write Last record */
	if( ( fwrite( &bidrec, sizeof(bidfwd), 1, wfptr ) == 0 ) ) {
		fprintf( stderr, "\nError: writing to %s\n", wf_new );
		fclose( wfptr );
		fclose( dirsys );
		return 1;
	}

        fclose( wfptr );

	/* replace original wfbid.sys with copy of wfbid.new */
	if( ! copy_ ( wf_new, wf_sys ) ){
		fprintf( stderr, "ERROR: Unable to copy '%s' to '%s'\n", wf_new, wf_sys );
		fprintf( stderr, "Please copy the file manually before starting XFBB\n" );
	}

	return 0;
}


int main(int argc, char *argv[])
{
	char Cmd[128];
	char *ap;
	int ch;
	int err = 0;
	int option_index = 0;
	static struct option long_options[] =
	{
	{ "help", 0, 0, 'h' },
	{ "verbose", 0, 0, 0 },
	{ "list", 0, 0, 0 },
	{ "rebuild", 0, 0, 0 },
	{ "search", 1, 0, 0 },
	{ "version", 0, 0, 'v' },
	{ 0, 0, 0, 0 }
	};
	int fbbpid;

	strcpy( Cmd, argv[0] );
	if( (ap = rindex( Cmd, '/') ) != NULL ) {
		ap++;
	}else{
		ap=Cmd;
	}

	while( (ch = getopt_long( argc, argv, "?hHvVlLsS", long_options, &option_index ) ) != (-1) ) {

		switch(ch){

			case 0:
				if( strcmp(long_options[option_index].name, "verbose" ) == 0 ){
					verbose=1;

				}else if( strcmp(long_options[option_index].name, "list" ) == 0 ){
					list=1;

				}else if( strcmp(long_options[option_index].name, "rebuild" ) == 0 ){
					rebuild=1;

				}else if( strcmp(long_options[option_index].name, "search" ) == 0 ){
					search=1;
					if( optarg ){
						strcpy( s_string, strupr(optarg) );
					}
				}

				break;

			case '?':
			case 'h':
			case 'H':
				do_usage( ap );
				return 1;

			case 'v':
			case 'V':
				do_head( ap );
				return 1;

			default:
				return 1;
		}
	}

	if( ReadInit() != 0 ){
		fprintf( stderr, "Error Processing Config file.\n" );
		return 1;
        }

	sprintf( dirmes_sys, "%s/dirmes.sys", data_path );
	sprintf( dirmes_new, "%s/dirmes.new", data_path );
	sprintf( dirmes_old, "%s/dirmes.old", data_path );

	sprintf( wf_sys, "%s/wfbid.sys", data_path );
	sprintf( wf_new, "%s/wfbid.new", data_path );
	sprintf( wf_old, "%s/wfbid.old", data_path );

	if( ! (list + rebuild + search) & ! (optind < argc) ){
		do_usage( ap );
		return 1;
	}

	do_head( ap );

	if (optind < argc) {
		while (optind < argc)
			err = err | do_bidlist(wf_sys, strupr(argv[optind++]));

		fprintf( stdout, "\n%ld Matches\n", match );
		fprintf( stdout, "of %ld Known Bids\n", good );
		return err;
	}

	if( search ){
		err = do_bidlist(wf_sys, s_string);
		fprintf( stdout, "\n%ld Matches\n", match );
		fprintf( stdout, "of %ld Known Bids\n", good );
		return err;
	}
	
	if( list ){
		err = do_bidlist(wf_sys, NULL);
		fprintf( stdout, "\n%ld Records\n", count );
		fprintf( stdout, "of %ld Known Bids\n", good );
		return err;
	}
	
	if( rebuild ){
		if( verbose ){
			fprintf( stdout, "Checking if XFBB is active.\n" );
		}

		fbbpid = Is_fbb();

		if( fbbpid == 127 ) {
			fprintf(stderr, "\007Error: Unable to determine XFBB is currently active.\n" );
			return 2;
		}

		if( fbbpid == 0 ) {
			fprintf(stderr, "\007Error: XFBB is currently active\n" );
			fprintf(stderr, "Cannot re-build wfbid.sys at this time.\n" );
			return 2;
		}

		err = rebuild_wfbids();
		if( err == 0 ){
			fprintf( stdout, "%ld Records Written\n", count );
			fprintf( stdout, "Done\n" );
		}
		return err;
	}
	
	return 0;
}
